from random import randint

import garaga.modulo_circuit_structs as structs
from garaga.definitions import CURVES, CurveID, G1Point, G2Point
from garaga.hints import neg_3
from garaga.hints.ecip import (
    n_coeffs_from_n_points,
    n_points_from_n_coeffs,
    slope_intercept,
)
from garaga.modulo_circuit import WriteOps
from garaga.modulo_circuit_structs import G1PointCircuit, G2PointCircuit, u384
from garaga.precompiled_circuits.cofactor_clearing import G1CofactorClearing
from garaga.precompiled_circuits.compilable_circuits.base import (
    BaseModuloCircuit,
    ModuloCircuit,
    PyFelt,
)
from garaga.precompiled_circuits.ec import (
    BasicEC,
    BasicECG2,
    ECIPCircuits,
    FakeGLVCircuits,
    IsOnCurveCircuit,
)


class DummyCircuit(BaseModuloCircuit):
    def __init__(
        self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0
    ) -> None:
        super().__init__(
            name="dummy",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        return [self.field(44), self.field(4)]

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ModuloCircuit(
            self.name,
            self.curve_id,
            generic_circuit=True,
            compilation_mode=self.compilation_mode,
        )
        x, y = circuit.write_elements(input, WriteOps.INPUT)
        r = circuit.sub(x, y)  # 40
        z = circuit.div(x, y)  # 4
        c = circuit.add(r, z)  # 44
        d = circuit.sub(r, z)  # 36
        e = circuit.mul(r, z)  # 120
        f = circuit.div(r, z)  # 8
        circuit.extend_output([r, z, c, d, e, f])

        return circuit


class IsOnCurveG1G2Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="is_on_curve_g1_g2",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        curve_id = CurveID(self.curve_id)
        order = CURVES[self.curve_id].n
        input = []
        n1, n2 = randint(1, order), randint(1, order)
        p1, p2 = G1Point.get_nG(curve_id, n1), G2Point.get_nG(curve_id, n2)
        pair = [p1.x, p1.y, p2.x[0], p2.x[1], p2.y[0], p2.y[1]]
        input.extend([self.field(x) for x in pair])
        input.append(self.field(CURVES[self.curve_id].a))
        input.append(self.field(CURVES[self.curve_id].b))
        input.append(self.field(CURVES[self.curve_id].b20))
        input.append(self.field(CURVES[self.curve_id].b21))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit: IsOnCurveCircuit = IsOnCurveCircuit(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        assert len(input) == 6 + 4, f"got {input} of len {len(input)}"
        p = circuit.write_struct(G1PointCircuit("p", input[0:2]), WriteOps.INPUT)
        q = circuit.write_struct(G2PointCircuit("q", input[2:6]), WriteOps.INPUT)
        a = circuit.write_struct(u384("a", [input[6]]), WriteOps.INPUT)
        b = circuit.write_struct(u384("b", [input[7]]), WriteOps.INPUT)
        b20 = circuit.write_struct(u384("b20", [input[8]]), WriteOps.INPUT)
        b21 = circuit.write_struct(u384("b21", [input[9]]), WriteOps.INPUT)

        circuit.set_consts(a, b, b20, b21)
        lhs, rhs = circuit._is_on_curve_G1(*p)
        lhs2, rhs2 = circuit._is_on_curve_G2(*q)
        zero_check = circuit.sub(lhs, rhs)
        zero_check_2 = [circuit.sub(lhs2[0], rhs2[0]), circuit.sub(lhs2[1], rhs2[1])]

        circuit.extend_struct_output(u384("zero_check_0", [zero_check]))
        circuit.extend_struct_output(u384("zero_check_1", [zero_check_2[0]]))
        circuit.extend_struct_output(u384("zero_check_2", [zero_check_2[1]]))

        return circuit


class IsOnCurveG1Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="is_on_curve_g1",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        random_point = G1Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(random_point.x))
        input.append(self.field(random_point.y))
        input.append(self.field(CURVES[self.curve_id].a))
        input.append(self.field(CURVES[self.curve_id].b))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicEC(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        px, py = circuit.write_struct(G1PointCircuit("p", input[0:2]), WriteOps.INPUT)
        a = circuit.write_struct(u384("a", [input[2]]), WriteOps.INPUT)
        b = circuit.write_struct(u384("b", [input[3]]), WriteOps.INPUT)

        lhs, rhs = circuit._is_on_curve_G1_weirstrass(px, py, a, b)
        zero_check = circuit.sub(lhs, rhs)
        circuit.extend_struct_output(u384("zero_check", [zero_check]))

        return circuit


class IsOnCurveG2Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="is_on_curve_g2",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        random_point = G2Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(random_point.x[0]))
        input.append(self.field(random_point.x[1]))
        input.append(self.field(random_point.y[0]))
        input.append(self.field(random_point.y[1]))
        input.append(self.field(CURVES[self.curve_id].a))
        input.append(self.field(CURVES[self.curve_id].b20))
        input.append(self.field(CURVES[self.curve_id].b21))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicECG2(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        x0, x1, y0, y1 = circuit.write_struct(
            G2PointCircuit("p", input[0:4]), WriteOps.INPUT
        )
        a = circuit.write_struct(u384("a", [input[4]]), WriteOps.INPUT)
        b20 = circuit.write_struct(u384("b20", [input[5]]), WriteOps.INPUT)
        b21 = circuit.write_struct(u384("b21", [input[6]]), WriteOps.INPUT)

        lhs, rhs = circuit._is_on_curve_G2_weirstrass(x0, x1, y0, y1, a, b20, b21)
        zero_check = [circuit.sub(lhs[0], rhs[0]), circuit.sub(lhs[1], rhs[1])]
        circuit.extend_struct_output(u384("zero_check_0", [zero_check[0]]))
        circuit.extend_struct_output(u384("zero_check_1", [zero_check[1]]))

        return circuit


class AddECPointsG2Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="add_ec_points_g2",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        P = G2Point.gen_random_point(CurveID(self.curve_id))
        Q = G2Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(P.x[0]))
        input.append(self.field(P.x[1]))
        input.append(self.field(P.y[0]))
        input.append(self.field(P.y[1]))
        input.append(self.field(Q.x[0]))
        input.append(self.field(Q.x[1]))
        input.append(self.field(Q.y[0]))
        input.append(self.field(Q.y[1]))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicECG2(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        px0, px1, py0, py1 = circuit.write_struct(
            G2PointCircuit("p", input[0:4]), WriteOps.INPUT
        )
        qx0, qx1, qy0, qy1 = circuit.write_struct(
            G2PointCircuit("q", input[4:8]), WriteOps.INPUT
        )

        (nx0, nx1), (ny0, ny1) = circuit.add_points(
            ((px0, px1), (py0, py1)), ((qx0, qx1), (qy0, qy1))
        )
        circuit.extend_struct_output(G2PointCircuit("result", [nx0, nx1, ny0, ny1]))

        return circuit


class DoubleECPointG2AEq0Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="double_ec_point_g2_a_eq_0",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        P = G2Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(P.x[0]))
        input.append(self.field(P.x[1]))
        input.append(self.field(P.y[0]))
        input.append(self.field(P.y[1]))

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicECG2(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        px0, px1, py0, py1 = circuit.write_struct(
            G2PointCircuit("p", input[0:4]), WriteOps.INPUT
        )

        (nx0, nx1), (ny0, ny1) = circuit.double_point_a_eq_0(((px0, px1), (py0, py1)))
        circuit.extend_struct_output(G2PointCircuit("result", [nx0, nx1, ny0, ny1]))

        return circuit


class SlopeInterceptSamePointCircuit(BaseModuloCircuit):
    def __init__(
        self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0
    ) -> None:
        super().__init__(
            name="slope_intercept_same_point",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        random_point = G1Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(random_point.x))
        input.append(self.field(random_point.y))
        input.append(self.field(CURVES[self.curve_id].a))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        px, py = circuit.write_struct(G1PointCircuit("p", input[0:2]), WriteOps.INPUT)
        a = circuit.write_struct(u384("a", [input[2]]), WriteOps.INPUT)
        m_A0, b_A0, xA0, yA0, xA2, yA2, coeff0, coeff2 = (
            circuit._slope_intercept_same_point((px, py), a)
        )
        if self.compilation_mode == 0:
            # In Cairo0, xA0 and yA0 are copied from the input
            circuit.extend_output([m_A0, b_A0, xA0, yA0, xA2, yA2, coeff0, coeff2])
        else:
            # In Cairo1, xA0 and yA0 are not copied from the input
            circuit.extend_struct_output(
                structs.SlopeInterceptOutput(
                    "mb", [m_A0, b_A0, xA2, yA2, coeff0, coeff2]
                )
            )

        return circuit


class AccumulateEvalPointChallengeSignedCircuit(BaseModuloCircuit):
    def __init__(
        self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0
    ) -> None:
        super().__init__(
            name="acc_eval_point_challenge_signed",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        A0 = G1Point.gen_random_point(CurveID(self.curve_id))
        m_A0, b_A0 = slope_intercept(A0, A0)
        input = []
        random_point = G1Point.gen_random_point(CurveID(self.curve_id))
        scalar = randint(0, 2**127)
        ep, en = neg_3.positive_negative_multiplicities(neg_3.neg_3_base_le(scalar))
        input.append(self.field(randint(0, CURVES[self.curve_id].p - 1)))
        input.append(self.field(m_A0.value))
        input.append(self.field(b_A0.value))
        input.append(self.field(A0.x))
        input.append(self.field(random_point.x))
        input.append(self.field(random_point.y))
        input.append(self.field(abs(ep)))
        input.append(self.field(abs(en)))
        input.append(self.field(1 if ep >= 0 else -1))
        input.append(self.field(1 if en >= 0 else -1))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        # acc, m, b, xA, px, py, ep, en, sp, sn = circuit.write_elements(
        #     input[0:10], WriteOps.INPUT
        # )
        acc, m, b, xA = (
            circuit.write_struct(u384("acc", [input[0]]), WriteOps.INPUT),
            circuit.write_struct(u384("m", [input[1]]), WriteOps.INPUT),
            circuit.write_struct(u384("b", [input[2]]), WriteOps.INPUT),
            circuit.write_struct(u384("xA", [input[3]]), WriteOps.INPUT),
        )
        px, py = circuit.write_struct(G1PointCircuit("p", input[4:6]), WriteOps.INPUT)
        ep = circuit.write_struct(u384("ep", [input[6]]), WriteOps.INPUT)
        en = circuit.write_struct(u384("en", [input[7]]), WriteOps.INPUT)
        sp = circuit.write_struct(u384("sp", [input[8]]), WriteOps.INPUT)
        sn = circuit.write_struct(u384("sn", [input[9]]), WriteOps.INPUT)

        res_acc = circuit._accumulate_eval_point_challenge_signed_same_point(
            acc, (m, b), xA, (px, py), ep, en, sp, sn
        )
        circuit.extend_struct_output(u384("res_acc", [res_acc]))

        return circuit


class RHSFinalizeAccCircuit(BaseModuloCircuit):
    def __init__(
        self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0
    ) -> None:
        super().__init__(
            name="rhs_finalize_acc",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        A0 = G1Point.gen_random_point(CurveID(self.curve_id))
        m_A0, b_A0 = slope_intercept(A0, A0)
        input = []
        Q = G1Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(randint(0, CURVES[self.curve_id].p - 1)))
        input.append(self.field(m_A0.value))
        input.append(self.field(b_A0.value))
        input.append(self.field(A0.x))
        input.append(self.field(Q.x))
        input.append(self.field(Q.y))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        # acc, m, b, xA, Qx, Qy = circuit.write_elements(input[0:6], WriteOps.INPUT)
        acc, m, b, xA = (
            circuit.write_struct(u384("acc", [input[0]]), WriteOps.INPUT),
            circuit.write_struct(u384("m", [input[1]]), WriteOps.INPUT),
            circuit.write_struct(u384("b", [input[2]]), WriteOps.INPUT),
            circuit.write_struct(u384("xA", [input[3]]), WriteOps.INPUT),
        )
        Qx, Qy = circuit.write_struct(
            G1PointCircuit("Q_result", input[4:6]), WriteOps.INPUT
        )
        res_acc = circuit._RHS_finalize_acc(acc, (m, b), xA, (Qx, Qy))
        circuit.extend_struct_output(u384("rhs", [res_acc]))

        return circuit


class EvalFunctionChallengeSingleCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        n_points: int = 1,
        auto_run: bool = True,
        compilation_mode: int = 0,
        batched: bool = False,
        generic_circuit: bool = True,
    ) -> None:
        self.n_points = n_points
        self.batched = batched
        self.generic_circuit = generic_circuit
        super().__init__(
            name=f"eval_fn_challenge_sing_{n_points}P" + ("_rlc" if batched else ""),
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        circuit = SlopeInterceptSamePointCircuit(self.curve_id, auto_run=False)
        xA, _yA, _A = circuit.build_input()
        m_A0, b_A0, xA0, yA0, xA2, yA2, coeff0, coeff2 = circuit._run_circuit_inner(
            [xA, _yA, _A]
        ).output
        input.extend([xA0.felt, _yA.felt, coeff0.felt])
        n_coeffs = n_coeffs_from_n_points(self.n_points, self.batched)
        for _ in range(sum(n_coeffs)):
            input.append(self.field(randint(0, CURVES[self.curve_id].p - 1)))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name,
            self.curve_id,
            compilation_mode=self.compilation_mode,
            generic_circuit=self.generic_circuit,
        )

        xA0, yA0 = circuit.write_struct(
            G1PointCircuit("A", [input[0], input[1]]), WriteOps.INPUT
        )
        coeff0 = circuit.write_struct(u384("coeff", [input[2]]), WriteOps.INPUT)

        all_coeffs = input[3:]

        def split_list(input_list, lengths):
            start_idx, result = 0, []
            for length in lengths:
                result.append(input_list[start_idx : start_idx + length])
                start_idx += length
            return result

        n_points = n_points_from_n_coeffs(len(all_coeffs), self.batched)
        _log_div_a_num, _log_div_a_den, _log_div_b_num, _log_div_b_den = split_list(
            all_coeffs, n_coeffs_from_n_points(n_points, self.batched)
        )
        log_div_a_num, log_div_a_den, log_div_b_num, log_div_b_den = (
            circuit.write_struct(
                structs.FunctionFeltCircuit(
                    name="SumDlogDiv" + ("Batched" if self.batched else ""),
                    elmts=[
                        structs.u384Span("log_div_a_num", _log_div_a_num),
                        structs.u384Span("log_div_a_den", _log_div_a_den),
                        structs.u384Span("log_div_b_num", _log_div_b_num),
                        structs.u384Span("log_div_b_den", _log_div_b_den),
                    ],
                ),
                WriteOps.INPUT,
            )
        )

        res = circuit._eval_function_challenge_single(
            (xA0, yA0),
            coeff0,
            log_div_a_num,
            log_div_a_den,
            log_div_b_num,
            log_div_b_den,
        )
        circuit.extend_struct_output(u384("res", [res]))

        return circuit


class EvalFunctionChallengeDuplCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        n_points: int = 1,
        auto_run: bool = True,
        compilation_mode: int = 0,
        batched: bool = False,
        generic_circuit: bool = True,
    ) -> None:
        self.n_points = n_points
        self.batched = batched
        self.generic_circuit = generic_circuit
        super().__init__(
            name=f"eval_fn_challenge_dupl_{n_points}P" + ("_rlc" if batched else ""),
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        circuit = SlopeInterceptSamePointCircuit(self.curve_id, auto_run=False)
        xA, _yA, _A = circuit.build_input()
        m_A0, b_A0, xA0, yA0, xA2, yA2, coeff0, coeff2 = circuit._run_circuit_inner(
            [xA, _yA, _A]
        ).output
        input.extend([xA0.felt, _yA.felt, xA2.felt, yA2.felt, coeff0.felt, coeff2.felt])
        n_coeffs = n_coeffs_from_n_points(self.n_points, self.batched)
        for _ in range(sum(n_coeffs)):
            input.append(self.field(randint(0, CURVES[self.curve_id].p - 1)))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name,
            self.curve_id,
            compilation_mode=self.compilation_mode,
            generic_circuit=self.generic_circuit,
        )

        xA0, yA0 = circuit.write_struct(
            G1PointCircuit("A0", [input[0], input[1]]), WriteOps.INPUT
        )
        xA2, yA2 = circuit.write_struct(
            G1PointCircuit("A2", [input[2], input[3]]), WriteOps.INPUT
        )
        coeff0 = circuit.write_struct(u384("coeff0", [input[4]]), WriteOps.INPUT)
        coeff2 = circuit.write_struct(u384("coeff2", [input[5]]), WriteOps.INPUT)

        all_coeffs = input[6:]

        def split_list(input_list, lengths):
            start_idx, result = 0, []
            for length in lengths:
                result.append(input_list[start_idx : start_idx + length])
                start_idx += length
            return result

        n_points = n_points_from_n_coeffs(len(all_coeffs), self.batched)
        _log_div_a_num, _log_div_a_den, _log_div_b_num, _log_div_b_den = split_list(
            all_coeffs, n_coeffs_from_n_points(n_points, self.batched)
        )
        log_div_a_num, log_div_a_den, log_div_b_num, log_div_b_den = (
            circuit.write_struct(
                structs.FunctionFeltCircuit(
                    name="SumDlogDiv" + ("Batched" if self.batched else ""),
                    elmts=[
                        structs.u384Span("log_div_a_num", _log_div_a_num),
                        structs.u384Span("log_div_a_den", _log_div_a_den),
                        structs.u384Span("log_div_b_num", _log_div_b_num),
                        structs.u384Span("log_div_b_den", _log_div_b_den),
                    ],
                ),
                WriteOps.INPUT,
            )
        )

        res = circuit._eval_function_challenge_dupl(
            (xA0, yA0),
            (xA2, yA2),
            coeff0,
            coeff2,
            log_div_a_num,
            log_div_a_den,
            log_div_b_num,
            log_div_b_den,
        )
        circuit.extend_struct_output(u384("res", [res]))

        return circuit


class InitFunctionChallengeDuplCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        n_points: int = 1,
        auto_run: bool = True,
        batched: bool = False,
        compilation_mode: int = 0,
    ) -> None:
        self.n_points = n_points
        self.batched = batched
        super().__init__(
            name=f"init_fn_challenge_dupl_{n_points}P" + ("_rlc" if batched else ""),
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random(), self.field.random()])  # xA0, xA2
        n_coeffs = n_coeffs_from_n_points(self.n_points, self.batched)
        for _ in range(sum(n_coeffs)):
            input.append(self.field(randint(0, CURVES[self.curve_id].p - 1)))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        xA0 = circuit.write_struct(u384("xA0", [input[0]]), WriteOps.INPUT)
        xA2 = circuit.write_struct(u384("xA2", [input[1]]), WriteOps.INPUT)

        all_coeffs = input[2:]

        def split_list(input_list, lengths):
            start_idx, result = 0, []
            for length in lengths:
                result.append(input_list[start_idx : start_idx + length])
                start_idx += length
            return result

        n_points = n_points_from_n_coeffs(len(all_coeffs), self.batched)
        _log_div_a_num, _log_div_a_den, _log_div_b_num, _log_div_b_den = split_list(
            all_coeffs, n_coeffs_from_n_points(n_points, self.batched)
        )

        log_div_a_num, log_div_a_den, log_div_b_num, log_div_b_den = (
            circuit.write_struct(
                structs.FunctionFeltCircuit(
                    name="SumDlogDiv",
                    elmts=[
                        structs.u384Span("log_div_a_num", _log_div_a_num),
                        structs.u384Span("log_div_a_den", _log_div_a_den),
                        structs.u384Span("log_div_b_num", _log_div_b_num),
                        structs.u384Span("log_div_b_den", _log_div_b_den),
                    ],
                ),
                WriteOps.INPUT,
            )
        )

        res = circuit._init_function_challenge_dupl(
            xA0, xA2, log_div_a_num, log_div_a_den, log_div_b_num, log_div_b_den
        )
        circuit.extend_struct_output(
            structs.FunctionFeltEvaluations("A0_evals", res[0:4])
        )
        circuit.extend_struct_output(
            structs.FunctionFeltEvaluations("A2_evals", res[4:8])
        )
        circuit.extend_struct_output(u384("xA0_power", [res[8]]))
        circuit.extend_struct_output(u384("xA2_power", [res[9]]))
        return circuit


class AccumulateFunctionChallengeDuplCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 0,
    ):
        super().__init__(
            name="acc_function_challenge_dupl",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend(
            [self.field.random() for _ in range(8)]
        )  # a_num, a_den, b_num, b_den accumulation evals * 2
        input.extend([self.field.random(), self.field.random()])  # xA0, xA2
        input.extend(
            [self.field.random(), self.field.random()]
        )  # xA0_power, xA2_power, corresponding the max degree of a_den or b_num of the previous accumulator
        input.extend(
            [self.field.random() for _ in range(4)]
        )  # next coefficients of a_num, a_den, b_num, b_den
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )

        f_a0_accs = circuit.write_struct(
            structs.FunctionFeltEvaluations("f_a0_accs", input[0:4]), WriteOps.INPUT
        )
        f_a1_accs = circuit.write_struct(
            structs.FunctionFeltEvaluations("f_a1_accs", input[4:8]), WriteOps.INPUT
        )
        xA0 = circuit.write_struct(u384("xA0", [input[8]]), WriteOps.INPUT)
        xA2 = circuit.write_struct(u384("xA2", [input[9]]), WriteOps.INPUT)
        xA0_power = circuit.write_struct(u384("xA0_power", [input[10]]), WriteOps.INPUT)
        xA2_power = circuit.write_struct(u384("xA2_power", [input[11]]), WriteOps.INPUT)
        next_a_num_coeff = circuit.write_struct(
            structs.GenericT("next_a_num_coeff", [input[12]]), WriteOps.INPUT
        )
        next_a_den_coeff = circuit.write_struct(
            structs.GenericT("next_a_den_coeff", [input[13]]), WriteOps.INPUT
        )
        next_b_num_coeff = circuit.write_struct(
            structs.GenericT("next_b_num_coeff", [input[14]]), WriteOps.INPUT
        )
        next_b_den_coeff = circuit.write_struct(
            structs.GenericT("next_b_den_coeff", [input[15]]), WriteOps.INPUT
        )

        res = circuit._accumulate_function_challenge_dupl(
            *f_a0_accs,
            *f_a1_accs,
            xA0,
            xA2,
            xA0_power,
            xA2_power,
            next_a_num_coeff,
            next_a_den_coeff,
            next_b_num_coeff,
            next_b_den_coeff,
        )

        circuit.extend_struct_output(
            structs.FunctionFeltEvaluations("next_f_a0_accs", res[0:4])
        )
        circuit.extend_struct_output(
            structs.FunctionFeltEvaluations("next_f_a1_accs", res[4:8])
        )
        circuit.extend_struct_output(u384("next_xA0_power", [res[8]]))
        circuit.extend_struct_output(u384("next_xA2_power", [res[9]]))
        return circuit


class FinalizeFunctionChallengeDuplCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 0,
    ):
        super().__init__(
            name="finalize_fn_challenge_dupl",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(8)])  # f_a0_accs, f_a1_accs
        input.extend([self.field.random() for _ in range(2)])  # yA0, yA2
        input.extend([self.field.random() for _ in range(2)])  # coeff_A0, coeff_A2
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ECIPCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        f_a0_accs = circuit.write_struct(
            structs.FunctionFeltEvaluations("f_a0_accs", input[0:4]), WriteOps.INPUT
        )
        f_a1_accs = circuit.write_struct(
            structs.FunctionFeltEvaluations("f_a1_accs", input[4:8]), WriteOps.INPUT
        )
        yA0 = circuit.write_struct(u384("yA0", [input[8]]), WriteOps.INPUT)
        yA2 = circuit.write_struct(u384("yA2", [input[9]]), WriteOps.INPUT)
        coeff_A0 = circuit.write_struct(u384("coeff_A0", [input[10]]), WriteOps.INPUT)
        coeff_A2 = circuit.write_struct(u384("coeff_A2", [input[11]]), WriteOps.INPUT)

        res = circuit._finalize_function_challenge_dupl(
            *f_a0_accs, *f_a1_accs, yA0, yA2, coeff_A0, coeff_A2
        )

        circuit.extend_struct_output(u384("res", [res]))
        return circuit


class AddECPointCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 0,
    ):
        super().__init__(
            name="add_ec_point",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        P = G1Point.gen_random_point(CurveID(self.curve_id))
        Q = G1Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(P.x))
        input.append(self.field(P.y))
        input.append(self.field(Q.x))
        input.append(self.field(Q.y))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicEC(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        circuit.generic_modulus = True
        xP, yP = circuit.write_struct(G1PointCircuit("p", input[0:2]), WriteOps.INPUT)
        xQ, yQ = circuit.write_struct(G1PointCircuit("q", input[2:4]), WriteOps.INPUT)
        xR, yR = circuit.add_points((xP, yP), (xQ, yQ))
        circuit.extend_struct_output(G1PointCircuit("r", [xR, yR]))

        return circuit


class DoubleECPointCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 0,
    ):
        super().__init__(
            name="double_ec_point",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        P = G1Point.gen_random_point(CurveID(self.curve_id))
        input.append(self.field(P.x))
        input.append(self.field(P.y))
        input.append(self.field(CURVES[self.curve_id].a))
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicEC(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        circuit.generic_modulus = True
        xP, yP = circuit.write_struct(G1PointCircuit("p", input[0:2]), WriteOps.INPUT)
        A = circuit.write_struct(u384("A_weirstrass", [input[2]]))
        xR, yR = circuit.double_point((xP, yP), A)
        circuit.extend_struct_output(G1PointCircuit("r", [xR, yR]))

        return circuit


class PrepareGLVFakeGLVPtsCircuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="prepare_glv_fake_glv_pts",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.append(self.field.random())  # Px
        input.append(self.field.random())  # P0y
        input.append(self.field.random())  # P1y
        input.append(self.field.random())  # Qx
        input.append(self.field.random())  # Q0y
        input.append(self.field.random())  # Phi_P0y
        input.append(self.field.random())  # Phi_P1y
        input.append(self.field.random())  # Phi_Q0y
        input.append(self.field.random())  # Gen_x
        input.append(self.field.random())  # Gen_y
        input.append(self.field.random())  # third_root

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = FakeGLVCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        circuit.generic_modulus = True
        Px = circuit.write_struct(u384("Px", [input[0]]), WriteOps.INPUT)
        P0y = circuit.write_struct(u384("P0y", [input[1]]), WriteOps.INPUT)
        P1y = circuit.write_struct(u384("P1y", [input[2]]), WriteOps.INPUT)
        Qx = circuit.write_struct(u384("Qx", [input[3]]), WriteOps.INPUT)
        Q0y = circuit.write_struct(u384("Q0y", [input[4]]), WriteOps.INPUT)
        Phi_P0y = circuit.write_struct(u384("Phi_P0y", [input[5]]), WriteOps.INPUT)
        Phi_P1y = circuit.write_struct(u384("Phi_P1y", [input[6]]), WriteOps.INPUT)
        Phi_Q0y = circuit.write_struct(u384("Phi_Q0y", [input[7]]), WriteOps.INPUT)
        Gen = circuit.write_struct(
            G1PointCircuit("Gen", [input[8], input[9]]), WriteOps.INPUT
        )
        third_root = circuit.write_struct(
            u384("third_root", [input[10]]), WriteOps.INPUT
        )

        (
            B1,
            B2,
            B3,
            B4,
            B5,
            B6,
            B7,
            B8,
            B9y,
            B10y,
            B11y,
            B12y,
            B13y,
            B14y,
            B15y,
            B16y,
            Phi_P0x,
            Phi_Q0x,
            Acc,
        ) = circuit.prepare_points_glv_fake_glv(
            Px, P0y, P1y, Qx, Q0y, Phi_P0y, Phi_P1y, Phi_Q0y, Gen, third_root
        )
        # circuit.exact_output_refs_needed = [B2[0], B3[0], B4[0], B5[0], B6[0], B7[0], B8[0], B10y, B11y, B12y, B13y, B14y, B15y, B16y, Acc[0], Acc[1]]
        circuit.extend_struct_output(G1PointCircuit("B1", [B1[0], B1[1]]))
        circuit.extend_struct_output(G1PointCircuit("B2", [B2[0], B2[1]]))
        circuit.extend_struct_output(G1PointCircuit("B3", [B3[0], B3[1]]))
        circuit.extend_struct_output(G1PointCircuit("B4", [B4[0], B4[1]]))
        circuit.extend_struct_output(G1PointCircuit("B5", [B5[0], B5[1]]))
        circuit.extend_struct_output(G1PointCircuit("B6", [B6[0], B6[1]]))
        circuit.extend_struct_output(G1PointCircuit("B7", [B7[0], B7[1]]))
        circuit.extend_struct_output(G1PointCircuit("B8", [B8[0], B8[1]]))
        circuit.extend_struct_output(u384("B9y", [B9y]))
        circuit.extend_struct_output(u384("B10y", [B10y]))
        circuit.extend_struct_output(u384("B11y", [B11y]))
        circuit.extend_struct_output(u384("B12y", [B12y]))
        circuit.extend_struct_output(u384("B13y", [B13y]))
        circuit.extend_struct_output(u384("B14y", [B14y]))
        circuit.extend_struct_output(u384("B15y", [B15y]))
        circuit.extend_struct_output(u384("B16y", [B16y]))
        circuit.extend_struct_output(u384("Phi_P0x", [Phi_P0x]))
        circuit.extend_struct_output(u384("Phi_Q0x", [Phi_Q0x]))
        circuit.extend_struct_output(G1PointCircuit("Acc", [Acc[0], Acc[1]]))
        return circuit


class PrepareFakeGLVPtsCircuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="prepare_fake_glv_pts",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.append(self.field.random())  # Px
        input.append(self.field.random())  # Py
        input.append(self.field.random())  # Qx
        input.append(self.field.random())  # Qy
        input.append(self.field.random())  # s2_sign
        input.append(self.field.random())  # A_weirstrass

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = FakeGLVCircuits(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        circuit.generic_modulus = True
        P = circuit.write_struct(
            G1PointCircuit("P", [input[0], input[1]]), WriteOps.INPUT
        )
        Q = circuit.write_struct(
            G1PointCircuit("Q", [input[2], input[3]]), WriteOps.INPUT
        )
        s2_sign = circuit.write_struct(u384("s2_sign", [input[4]]), WriteOps.INPUT)
        A_weirstrass = circuit.write_struct(
            u384("A_weirstrass", [input[5]]), WriteOps.INPUT
        )

        (
            T1,
            T2,
            T3,
            T4,
            T5y,
            T6y,
            T7y,
            T8y,
            T9,
            T10,
            T11,
            T12,
            T13y,
            T14y,
            T15y,
            T16y,
            R2,
            R0y,
        ) = circuit.prepare_points_fake_glv(P, Q, s2_sign, A_weirstrass)
        # circuit.exact_output_refs_needed = [B2[0], B3[0], B4[0], B5[0], B6[0], B7[0], B8[0], B10y, B11y, B12y, B13y, B14y, B15y, B16y, Acc[0], Acc[1]]
        circuit.extend_struct_output(G1PointCircuit("T1", [T1[0], T1[1]]))
        circuit.extend_struct_output(G1PointCircuit("T2", [T2[0], T2[1]]))
        circuit.extend_struct_output(G1PointCircuit("T3", [T3[0], T3[1]]))
        circuit.extend_struct_output(G1PointCircuit("T4", [T4[0], T4[1]]))
        circuit.extend_struct_output(u384("T5y", [T5y]))
        circuit.extend_struct_output(u384("T6y", [T6y]))
        circuit.extend_struct_output(u384("T7y", [T7y]))
        circuit.extend_struct_output(u384("T8y", [T8y]))
        circuit.extend_struct_output(G1PointCircuit("T9", [T9[0], T9[1]]))
        circuit.extend_struct_output(G1PointCircuit("T10", [T10[0], T10[1]]))
        circuit.extend_struct_output(G1PointCircuit("T11", [T11[0], T11[1]]))
        circuit.extend_struct_output(G1PointCircuit("T12", [T12[0], T12[1]]))
        circuit.extend_struct_output(u384("T13y", [T13y]))
        circuit.extend_struct_output(u384("T14y", [T14y]))
        circuit.extend_struct_output(u384("T15y", [T15y]))
        circuit.extend_struct_output(u384("T16y", [T16y]))
        circuit.extend_struct_output(G1PointCircuit("R2", [R2[0], R2[1]]))
        circuit.extend_struct_output(u384("R0y", [R0y]))
        return circuit


class QuadrupleAndAdd9Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="quadruple_and_add_9",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.append(self.field.random())  # Px
        input.append(self.field.random())  # Py
        for _ in range(9):
            input.append(self.field.random())  # Qx
            input.append(self.field.random())  # Qy
        input.append(self.field.random())  # A_weirstrass

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = BasicEC(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        circuit.generic_modulus = True
        P = circuit.write_struct(
            G1PointCircuit("P", [input[0], input[1]]), WriteOps.INPUT
        )
        Qs = []
        for i in range(2, len(input) - 1, 2):
            Qs.append(
                circuit.write_struct(
                    G1PointCircuit(f"Q_{i//2}", [input[i], input[i + 1]]),
                    WriteOps.INPUT,
                )
            )

        A_weirstrass = circuit.write_struct(
            u384("A_weirstrass", [input[len(input) - 1]]), WriteOps.INPUT
        )

        Rx, Ry = circuit.n_quadruple_and_add(P, Qs, A_weirstrass)
        circuit.extend_struct_output(G1PointCircuit("R", [Rx, Ry]))
        return circuit


class ClearCofactorBLS12_381Circuit(BaseModuloCircuit):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 0):
        super().__init__(
            name="clear_cofactor_bls12_381",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.append(self.field.random())  # Px
        input.append(self.field.random())  # Py
        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = G1CofactorClearing(
            self.name, self.curve_id, compilation_mode=self.compilation_mode
        )
        circuit.generic_modulus = True
        P = circuit.write_struct(
            G1PointCircuit("P", [input[0], input[1]]), WriteOps.INPUT
        )

        res = circuit.clear_cofactor(P)
        circuit.extend_struct_output(G1PointCircuit("res", [res[0], res[1]]))
        return circuit



import hashlib
import secrets
from dataclasses import dataclass

from garaga.definitions import CURVES, CurveID, G1G2Pair, G1Point, G2Point
from garaga.drand.client import DrandNetwork, digest_func
from garaga.hints.tower_backup import E12
from garaga.signature import hash_to_curve


@dataclass
class CipherText:
    U: G2Point
    V: bytes
    W: bytes

    def __post_init__(self):
        assert len(self.V) == len(self.W) == 16

    def serialize_to_cairo(self):
        code = f"""CipherText {{
        U: {self.U.serialize_to_cairo(name="U", raw=True)},
        V: [{', '.join(f'0x{b:02x}' for b in self.V)}],
        W: [{', '.join(f'0x{b:02x}' for b in self.W)}],
        }}
        """
        return code


def encrypt_for_round(
    drand_public_key: G2Point, round: int, message: bytes, debug: bool = False
) -> CipherText:
    assert len(message) == 16, f"Message should be 16 bytes: {len(message)}"

    msg_at_round = digest_func(round)
    # print(f"msg_at_round list of ints: {list(msg_at_round)}")
    pt_at_round = hash_to_curve(msg_at_round, CurveID.BLS12_381)

    gid: E12 = G1G2Pair.pair([G1G2Pair(p=pt_at_round, q=drand_public_key)])

    if debug:
        # Use a fixed sigma for debugging:
        sigma = b"0000000000000000"
    else:
        sigma: bytes = secrets.token_bytes(16)

    assert len(sigma) == 16

    hasher = hashlib.sha256()
    hasher.update(b"IBE-H3")
    hasher.update(sigma)
    hasher.update(message)
    r = int.from_bytes(expand_message_drand(hasher.digest(), 32), "little")
    # print(f"r list of ints: {r}")
    r = r % CURVES[CurveID.BLS12_381.value].n

    # U = r * G2
    U = G2Point.get_nG(CurveID.BLS12_381, r)

    # print(f"U: {U}")
    # V = sigma XOR H (rGid)
    rGid: E12 = gid**r

    # print(f"rGid: {rGid.value_coeffs}")

    rgid_serialized = rGid.serialize()
    # Print bytes as hex strings:
    # print(f"rGid hex : {rgid_serialized.hex()}")
    # print(f"rGid serialized: {list(rgid_serialized)} len: {len(rgid_serialized)}")
    rgid_hash = hashlib.sha256()
    rgid_hash.update(b"IBE-H2")
    rgid_hash.update(rgid_serialized)
    rgid_hash = rgid_hash.digest()
    # Take first 16 bytes only :
    rgid_hash = rgid_hash[:16]
    # print(f"rgid_hash hex: {rgid_hash.hex()}")

    V = bytes([a ^ b for a, b in zip(sigma, rgid_hash)])
    # print(f"V hex: {V.hex()}")

    # 6. Compute W = M XOR H(sigma)
    W = bytes([a ^ b for a, b in zip(message, rgid_hash)])
    sigma_hash = hashlib.sha256()
    sigma_hash.update(b"IBE-H4")
    sigma_hash.update(sigma)
    sigma_hash = sigma_hash.digest()[:16]  # First 16 bytes only
    # print(f"sigma_hash hex: {sigma_hash.hex()}")
    W = bytes([a ^ b for a, b in zip(message, sigma_hash)])
    # print(f"W hex: {W.hex()}")

    return CipherText(U, V, W)


def decrypt_at_round(signature_at_round: G1Point, c: CipherText):
    # 1. Compute sigma = V XOR H2(e(rP,Sig))

    r_gid: E12 = G1G2Pair.pair([G1G2Pair(p=signature_at_round, q=c.U)])
    # for i, v in enumerate(r_gid.value_coeffs):
    #     print(f"rgid_{i}: {io.int_to_u384(v, as_hex=False)}")
    r_gid_serialized = r_gid.serialize()
    rgid_hash = hashlib.sha256()

    pre_image = bytearray(b"IBE-H2")
    pre_image.extend(r_gid_serialized)
    rgid_hash.update(pre_image)

    # for i in range(0, len(pre_image), 4):
    #     print(f"pre_image[{i}:{i+4}]: {pre_image[i:i+4].hex()}")

    rgid_hash = rgid_hash.digest()

    rgid_hash = rgid_hash[:16]  # First 16 bytes only

    sigma = bytes([a ^ b for a, b in zip(c.V, rgid_hash)])

    # print(f"sigma hex: {sigma.hex()}")

    # 2. Compute Msg = W XOR H4(sigma)
    sigma_hash = hashlib.sha256()
    sigma_hash.update(b"IBE-H4")
    sigma_hash.update(sigma)
    sigma_hash = sigma_hash.digest()[:16]  # First 16 bytes only

    message = bytes([a ^ b for a, b in zip(c.W, sigma_hash)])
    # print(f"message utf-8: {message.decode('utf-8')}")

    # 3. Check U = G^r

    rh = hashlib.sha256()
    rh.update(b"IBE-H3")
    rh.update(sigma)
    rh.update(message)
    rh = rh.digest()
    rh = expand_message_drand(rh, 32)

    r = int.from_bytes(rh, "little")
    r = r % CURVES[CurveID.BLS12_381.value].n
    U = G2Point.get_nG(CurveID.BLS12_381, r)
    assert U == c.U
    return message


def expand_message_drand(msg: bytes, buf_size: int) -> bytes:
    BITS_TO_MASK_FOR_BLS12381 = 1
    order = CURVES[CurveID.BLS12_381.value].n
    for i in range(1, 65536):  # u16::MAX is 65535
        # Hash iteratively: H(i || msg)
        h = hashlib.sha256()
        pre_image = bytearray(i.to_bytes(2, byteorder="little"))
        pre_image.extend(msg)
        h.update(pre_image)
        hash_result = bytearray(h.digest())
        # Mask the first byte
        hash_result[0] >>= BITS_TO_MASK_FOR_BLS12381

        reversed_hash = hash_result[::-1]

        scalar = int.from_bytes(reversed_hash, "little") % order
        if scalar != 0:
            return reversed_hash[:buf_size]

    raise ValueError(
        "You are insanely unlucky and should have been hit by a meteor before now"
    )


def write_cairo1_test(msg: bytes, round: int, network: DrandNetwork):
    chain_infos = print_all_chain_info()
    chain = chain_infos[network]
    master = chain.public_key
    ciphertext: CipherText = encrypt_for_round(master, round, msg, debug=True)

    signature_at_round = get_randomness(chain.hash, round).signature_point
    _ = decrypt_at_round(signature_at_round, ciphertext)

    comment_with_params_used = f"""
    // msg: {msg}
    // round: {round}
    // network: {network}
    """
    code = f"""
    #[test]
    fn test_decrypt_at_round() {{
    {comment_with_params_used}
    {signature_at_round.serialize_to_cairo(name="signature_at_round")}
    let ciph = {ciphertext.serialize_to_cairo()};
    let msg_decrypted = decrypt_at_round(signature_at_round, ciph);
    assert(msg_decrypted.span() == [{', '.join(f'0x{b:02x}' for b in msg)}].span(), 'wrong msg');
    }}
    """
    return code


if __name__ == "__main__":
    from garaga.drand.client import DrandNetwork, get_randomness, print_all_chain_info

    # chain_infos = print_all_chain_info()
    network = DrandNetwork.quicknet
    # chain = chain_infos[network]

    # master = chain.public_key

    round = 128

    msg = b"hello\x00\x00\x00\x00\x00\x00\x00\x00abc"
    # ciph = encrypt_for_round(master, round, msg)

    # # print(f"CipherText: {ciph}")

    # # print(f"V : {list(ciph.V)}")
    # # print(f"W : {list(ciph.W)}")

    # chain = chain_infos[network]
    # beacon = get_randomness(chain.hash, round)
    # signature_at_round = beacon.signature_point

    # print(f"signature_at_round: {signature_at_round}")
    # msg_decrypted = decrypt_at_round(signature_at_round, ciph)
    # assert msg_decrypted == msg

    # print(f"msg utf-8: {msg.decode('utf-8')}")
    # print(f"msg_decrypted utf-8: {msg_decrypted.decode('utf-8')}")

    code = write_cairo1_test(msg, round, network)

    print(code)



import subprocess
from pathlib import Path
from typing import Annotated

import typer
from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.tree import Tree

from garaga.definitions import ProofSystem
from garaga.starknet.cli.utils import (
    complete_proof_system,
    get_default_vk_path,
    get_package_version,
)
from garaga.starknet.groth16_contract_generator.generator import (
    ECIP_OPS_CLASS_HASH,
    gen_groth16_verifier,
)
from garaga.starknet.honk_contract_generator.generator_honk import (
    BB_VERSION,
    BBUP_VERSION,
    NARGO_VERSION,
    gen_honk_verifier,
)


def check_nargo_version(version: str) -> tuple[bool, str]:
    return check_version("nargo", version)


def check_bb_version(version: str) -> tuple[bool, str]:
    return check_version("bb", version)


def check_version(cmd: str, version: str) -> tuple[bool, str]:
    result = subprocess.run(
        [cmd, "--version"], capture_output=True, text=True, check=True
    )
    result_version = result.stdout.strip()
    return (
        result.returncode == 0 and version in result_version,
        result_version,
    )


def _get_bb_oracle_hash_options() -> list[str]:
    """Get the available oracle hash options from bb command"""
    result_hash = subprocess.run(
        ["bb", "prove", "-h"], capture_output=True, text=True, check=True
    )
    # Extract oracle hash options
    oracle_hash_options = []
    lines = result_hash.stdout.splitlines()
    for i, line in enumerate(lines):
        if "--oracle_hash" in line:
            # Look in the next few lines for the options
            for j in range(i + 1, min(i + 100, len(lines))):
                if "Options: {" in lines[j]:
                    options_str = lines[j].split("{")[1].split("}")[0]
                    oracle_hash_options = [
                        opt.strip() for opt in options_str.split(",")
                    ]
                    break
            break

    return oracle_hash_options


def check_bb_starknet_support(
    version: str, with_starknet: bool = False
) -> tuple[bool, str, bool]:
    """Check if bb supports starknet and meets the version requirement"""
    version_ok, version_str = check_bb_version(version)
    oracle_options = _get_bb_oracle_hash_options()
    has_starknet = "starknet" in oracle_options
    if with_starknet:
        return (version_ok and has_starknet, version_str, has_starknet)
    else:
        return (version_ok, version_str, has_starknet)


def print_bb_warning(soft: bool = False, with_starknet: bool = False):
    bb_version_ok, bb_version, has_starknet = check_bb_starknet_support(
        BB_VERSION, with_starknet
    )
    nargo_version_ok, nargo_version = check_version("nargo", NARGO_VERSION)
    if not bb_version_ok or not nargo_version_ok:
        console = Console()
        console.print(
            Panel(
                f"[bold]Detected versions of bb and nargo are not compatible with the required versions by garaga {get_package_version()}.[/bold]\n"
                + (f"This is the most likely cause of the error." if not soft else f""),
                title="[bold red]Warning[/bold red]",
                border_style="red",
            )
        )
        console.print(
            f"Detected bb version: [bold red]{bb_version}[/bold red] with support for starknet optimized oracle hash: [bold red]{has_starknet}[/bold red]\n"
            f"Detected nargo version: [bold red]{nargo_version}[/bold red]"
        )
        console.print(
            Panel(
                "[bold]Please ensure you have the expected versions of bb and nargo installed for garaga to work correctly.[/bold]\n"
                + "You can install the required versions using the following commands:\n"
                + f"[bold yellow]bbup --version {BBUP_VERSION}[/bold yellow]\n"
                + f"[bold yellow]noirup --version {NARGO_VERSION}[/bold yellow]\n"
                + (
                    f"Check [blue]https://noir-lang.org/docs/getting_started/quick_start[/blue] for more information.\n"
                    if not soft
                    else ""
                )
                + f"Please ensure the verifying key was generated with the correct versions of bb and nargo before generating the verifier.",
                border_style="yellow",
                title="[bold yellow]Fixing the issue[/bold yellow]",
            )
        )
    else:
        print(
            f"The detected versions of bb and nargo are correctly aligned with garaga {get_package_version()}. "
        )


def gen(
    system: Annotated[
        ProofSystem,
        typer.Option(help="Proof system", autocompletion=complete_proof_system),
    ],
    project_name: Annotated[
        str,
        typer.Option(
            help="The verifer name (used for the folder name, scarb project name, and contract name)",
            prompt="\033[1;35mPlease enter the name of your project. Press enter for default name.\033[0m",
            default_factory=lambda: "my_project",
        ),
    ],
    vk: Annotated[
        Path,
        typer.Option(
            help="Path to the verification key file. Expects a JSON for groth16, binary format for Honk. Must be provided, except when the system supports an universal key (ex: Risc0)",
            file_okay=True,
            dir_okay=False,
            exists=True,
            autocompletion=lambda: [],  # See https://github.com/fastapi/typer/discussions/731
        ),
    ] = None,
):
    """
    Generate a Cairo verifier for a given proof system.
    Automatically detects the curve from the verification key.
    """
    cwd = Path.cwd()
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=False,
        console=Console(color_system="truecolor"),
    ) as progress:
        progress.add_task(
            f"[bold cyan]Generating Smart Contract project for [bold yellow]{system}[/bold yellow] using [bold yellow]{Path(vk).name}[/bold yellow]...[/bold cyan]",
            total=None,
        )
        if vk == None:
            vk = get_default_vk_path(system)
        match system:
            case ProofSystem.Groth16 | ProofSystem.Risc0Groth16:
                gen_groth16_verifier(
                    vk=vk,
                    output_folder_path=cwd,
                    output_folder_name=project_name,
                    ecip_class_hash=ECIP_OPS_CLASS_HASH,
                    cli_mode=True,
                )
            case (
                ProofSystem.UltraKeccakHonk
                | ProofSystem.UltraStarknetHonk
                | ProofSystem.UltraKeccakZKHonk
                | ProofSystem.UltraStarknetZKHonk
            ):
                try:
                    gen_honk_verifier(
                        vk=vk,
                        output_folder_path=cwd,
                        output_folder_name=project_name,
                        system=system,
                        cli_mode=True,
                    )
                except Exception as e:
                    import traceback

                    print(
                        f"\n[bold red]Error:[/bold red] An error occurred while generating the verifier: \n{traceback.format_exc()}"
                    )
                    print_bb_warning(
                        with_starknet=system == ProofSystem.UltraStarknetHonk
                        or system == ProofSystem.UltraStarknetZKHonk
                    )

                    raise typer.Exit(code=1)

            case _:
                raise ValueError(f"Unsupported proof system: {system}")

    print("[bold green]Done![/bold green]")
    print("[bold cyan]Smart Contract project created:[/bold cyan]")

    def generate_tree(path: Path, tree: Tree) -> None:
        for entry in sorted(path.iterdir(), key=lambda e: e.name):
            if entry.is_dir():
                branch = tree.add(f"[bold blue]{entry.name}/[/bold blue]")
                generate_tree(entry, branch)
            else:
                tree.add(f"[green]{entry.name}[/green]")

    root = Tree(f"[bold yellow]{cwd}/{project_name}/[/bold yellow]")
    generate_tree(cwd / project_name, root)
    print(root)

    print(
        f"[bold]You can now test the main endpoint of the verifier using a proof and `garaga calldata` command.[/bold]"
    )
    if system in [
        ProofSystem.UltraKeccakHonk,
        ProofSystem.UltraStarknetHonk,
        ProofSystem.UltraKeccakZKHonk,
        ProofSystem.UltraStarknetZKHonk,
    ]:
        if (
            system == ProofSystem.UltraStarknetHonk
            or system == ProofSystem.UltraStarknetZKHonk
        ):
            print_bb_warning(soft=True, with_starknet=True)
        else:
            print_bb_warning(soft=True)



from __future__ import annotations

import random
from dataclasses import dataclass
from typing import Generic, TypeVar

from sympy import legendre_symbol, sqrt_mod

T = TypeVar("T", "PyFelt", "Fp2")


@dataclass(slots=True)
class PyFelt:
    """
    Represents an element of a finite field. The dataclass decorator automatically provides implementations for __init__, __repr__, __eq__, and __hash__ methods.
    Supports the following operations through magic methods:
    - Addition ((PyFelt | int) + (PyFelt | int))
    - Subtraction ((PyFelt | int) - (PyFelt | int))
    - Multiplication ((PyFelt | int) * (PyFelt | int))
    - Division (PyFelt / PyFelt)
    - Negation (-PyFelt)
    - Inversion (inv(PyFelt))
    - Exponentiation (PyFelt ** int)
    - Equality (PyFelt == PyFelt)
    - Inequality (PyFelt != PyFelt)
    """

    value: int
    p: int

    @property
    def felt(self) -> PyFelt:
        return self

    def __repr__(self) -> str:
        p_str = f"0x{self.p:0x}"
        if len(p_str) > 10:
            p_str = f"{p_str[:6]}...{p_str[-4:]}"
        return f"PyFelt({self.value}, {p_str})"

    def __add__(self, right: PyFelt | int) -> PyFelt:
        p = self.p
        if isinstance(right, PyFelt):
            return PyFelt((self.value + right.value) % p, p)
        if isinstance(right, int):
            return PyFelt((self.value + right) % p, p)
        raise TypeError(f"Cannot add PyFelt and {type(right)}")

    def __neg__(self) -> PyFelt:
        p = self.p
        return PyFelt((-self.value) % p, p)

    def __sub__(self, right: PyFelt | int) -> PyFelt:
        p = self.p
        if isinstance(right, PyFelt):
            return PyFelt((self.value - right.value) % p, p)
        if isinstance(right, int):
            return PyFelt((self.value - right) % p, p)
        raise TypeError(f"Cannot subtract PyFelt and {type(right)}")

    def __mul__(self, right: PyFelt | int) -> PyFelt:
        p = self.p
        if isinstance(right, PyFelt):
            return PyFelt((self.value * right.value) % p, p)
        if isinstance(right, int):
            return PyFelt((self.value * right) % p, p)
        raise TypeError(f"Cannot multiply PyFelt and {type(right)}")

    def __rmul__(self, left: PyFelt | int) -> PyFelt:
        return self.__mul__(left)

    def __inv__(self) -> PyFelt:
        try:
            inv = pow(self.value, -1, self.p)
        except ValueError:
            raise ValueError(f"Cannot invert {self.value} modulo {self.p}")
        return PyFelt(inv, self.p)

    def __truediv__(self, right: PyFelt) -> PyFelt:
        assert isinstance(self, PyFelt) and isinstance(
            right, PyFelt
        ), f"Cannot divide {type(self)} by {type(right)}"
        return self * right.__inv__()

    def __pow__(self, exponent: int) -> PyFelt:
        return PyFelt(pow(self.value, exponent, self.p), self.p)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, PyFelt):
            return self.value == other.value
        if isinstance(other, int):
            return self.value == other
        raise TypeError(f"Cannot compare PyFelt and {type(other)}")

    def __lt__(self, other: PyFelt | int) -> bool:
        if isinstance(other, PyFelt):
            return self.value < other.value
        if isinstance(other, int):
            return self.value < other
        raise TypeError(f"Cannot compare PyFelt and {type(other)}")

    def __le__(self, other: PyFelt | int) -> bool:
        if isinstance(other, PyFelt):
            return self.value <= other.value
        if isinstance(other, int):
            return self.value <= other
        raise TypeError(f"Cannot compare PyFelt and {type(other)}")

    def __gt__(self, other: PyFelt | int) -> bool:
        if isinstance(other, PyFelt):
            return self.value > other.value
        if isinstance(other, int):
            return self.value > other
        raise TypeError(f"Cannot compare PyFelt and {type(other)}")

    def __ge__(self, other: PyFelt | int) -> bool:
        if isinstance(other, PyFelt):
            return self.value >= other.value
        if isinstance(other, int):
            return self.value >= other
        raise TypeError(f"Cannot compare PyFelt and {type(other)}")

    def __rlt__(self, left: int) -> bool:
        return left < self.value

    def __rle__(self, left: int) -> bool:
        return left <= self.value

    def __rgt__(self, left: int) -> bool:
        return left > self.value

    def __rge__(self, left: int) -> bool:
        return left >= self.value

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __radd__(self, left: PyFelt | int) -> PyFelt:
        return self.__add__(left)

    def __rsub__(self, left: PyFelt | int) -> PyFelt:
        return -self.__sub__(left)

    def __rtruediv__(self, left: PyFelt | int) -> PyFelt:
        return self.__inv__().__mul__(left)

    def is_quad_residue(self) -> bool:
        if self.value == 0:
            return True
        return legendre_symbol(self.value, self.p) == 1

    def sqrt(self, min_root: bool = True) -> PyFelt:
        if not self.is_quad_residue():
            raise ValueError("Cannot square root a non-quadratic residue")
        roots = sqrt_mod(self.value, self.p, all_roots=True)
        if min_root:
            return PyFelt(min(roots), self.p)
        else:
            return PyFelt(max(roots), self.p)


@dataclass(slots=True)
class Fp2:
    a0: PyFelt
    a1: PyFelt

    def __post_init__(self):
        assert self.a0.p == self.a1.p, "Fields must be the same"

    @property
    def p(self) -> int:
        return self.a0.p

    @property
    def value(self) -> tuple[int, int]:
        return self.a0.value, self.a1.value

    @staticmethod
    def random(p: int, max_value: int = None) -> Fp2:
        if max_value is None:
            max_value = p - 1

        rnd1 = random.randint(0, max_value)
        rnd2 = random.randint(0, max_value)
        return Fp2(
            PyFelt(rnd1, p),
            PyFelt(rnd2, p),
        )

    @staticmethod
    def one(p: int) -> Fp2:
        return Fp2(PyFelt(1, p), PyFelt(0, p))

    @staticmethod
    def zero(p: int) -> Fp2:
        return Fp2(PyFelt(0, p), PyFelt(0, p))

    def __repr__(self) -> str:
        return f"Fp2({self.a0}, {self.a1})"

    def __add__(self, other: Fp2) -> Fp2:
        if isinstance(other, Fp2):
            return Fp2(self.a0 + other.a0, self.a1 + other.a1)
        else:
            raise TypeError(f"Cannot add Fp2 and {type(other)}")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Fp2):
            return self.a0 == other.a0 and self.a1 == other.a1 and self.p == other.p
        else:
            raise TypeError(f"Cannot compare Fp2 and {type(other)}")

    def __neg__(self) -> Fp2:
        return Fp2(-self.a0, -self.a1)

    def __sub__(self, other: Fp2) -> Fp2:
        return self.__add__(-other)

    def __mul__(self, other: Fp2 | PyFelt | int) -> Fp2:
        if isinstance(other, PyFelt):
            assert other.p == self.a0.p, "Fields must be the same"
            return Fp2(self.a0 * other, self.a1 * other)
        elif isinstance(other, int):
            return Fp2(self.a0 * other, self.a1 * other)
        elif isinstance(other, Fp2):
            # (a0 + a1 * i) * (b0 + b1 * i) = a0 * b0 - a1 * b1 + (a0 * b1 + a1 * b0) * i
            return Fp2(
                self.a0 * other.a0 - self.a1 * other.a1,
                self.a0 * other.a1 + self.a1 * other.a0,
            )
        else:
            raise TypeError(f"Cannot multiply Fp2 and {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Fp2):
            return self * other.__inv__()
        elif isinstance(other, int):
            return self * pow(other, -1, self.p)

        return NotImplemented

    def __rtruediv__(self, other) -> Fp2:
        if isinstance(other, Fp2):
            return other * self.__inv__()
        elif isinstance(other, int):
            return other * self.__inv__()

        return NotImplemented

    def __inv__(self) -> Fp2:
        t0, t1 = (self.a0 * self.a0, self.a1 * self.a1)
        t0 = t0 + t1
        t1 = pow(t0.value, -1, self.p)
        return Fp2(self.a0 * t1, -(self.a1 * t1))

    def __pow__(self, p: int) -> Fp2:
        """
        Compute x**p in F_p^2 using square-and-multiply algorithm.
        Args:
        p: The exponent, a non-negative integer.
        Returns:
        x**p in F_p^2, represented similarly as x.
        """
        assert isinstance(p, int) and p >= 0

        # Handle the easy cases.
        if p == 0:
            # x**0 = 1, where 1 is the multiplicative identity in F_p^2.
            return Fp2(PyFelt(1, self.p), PyFelt(0, self.p))
        elif p == 1:
            # x**1 = x.
            return self

        # Start the computation.
        result = self.one(
            self.p
        )  # Initialize result as the multiplicative identity in F_p^2.
        temp = self  # Initialize temp as self.

        # Loop through each bit of the exponent p.
        for bit in reversed(bin(p)[2:]):  # [2:] to strip the "0b" prefix.
            if bit == "1":
                result = result * temp
            temp = temp * temp

        return result

    def norm(self) -> PyFelt:
        return self.a0 * self.a0 + self.a1 * self.a1

    def legendre(self) -> int:
        norm = self.norm()
        return legendre_symbol(norm.value, self.p)

    def is_quad_residue(self) -> bool:
        return self.legendre() == 1

    def sqrt(self) -> Fp2:
        if not self.is_quad_residue():
            raise ValueError("Cannot square root a non-quadratic residue")
        assert self.p % 4 == 3, "p must be 3 mod 4 to use this sqrt"
        min_one = Fp2(PyFelt(-1 % self.p, self.p), PyFelt(0, self.p))

        a = self
        a1 = a ** ((self.p - 3) // 4)
        alpha = a1 * a1 * a
        a0 = alpha**self.p * alpha
        if a0 == min_one:
            return ValueError("Cannot square root a non-quadratic residue")

        x0 = a1 * a
        if alpha == min_one:
            i = Fp2(PyFelt(0, self.p), PyFelt(1, self.p))
            x = i * x0
        else:
            b = (Fp2.one(self.p) + alpha) ** ((self.p - 1) // 2)
            x = b * x0

        # Return the root as is, without forcing a specific sign
        return x

    def lexicographically_largest(self) -> bool:
        """Check if this Fp2 element is lexicographically largest."""
        if self.a1.value > (self.p - 1) // 2:
            return True
        if self.a1.value < (self.p - 1) // 2:
            return False
        return self.a0.value > (self.p - 1) // 2


@dataclass(slots=True)
class BaseField:
    p: int

    def __call__(self, integer: int) -> PyFelt:
        return PyFelt(integer % self.p, self.p)

    def zero(self) -> PyFelt:
        return PyFelt(0, self.p)

    def one(self) -> PyFelt:
        return PyFelt(1, self.p)

    def random(self, max_value: int = None) -> PyFelt:
        if max_value is None:
            max_value = self.p - 1
        return PyFelt(random.randint(0, max_value), self.p)

    @property
    def type(self) -> type[PyFelt]:
        return PyFelt


@dataclass(slots=True)
class BaseFp2Field:
    p: int

    def __call__(self, a: tuple[int, int] | int) -> Fp2:
        if isinstance(a, tuple):
            a0, a1 = a
        else:
            a0, a1 = a, 0
        return Fp2(PyFelt(a0 % self.p, self.p), PyFelt(a1 % self.p, self.p))

    def zero(self) -> Fp2:
        return Fp2(PyFelt(0, self.p), PyFelt(0, self.p))

    def one(self) -> Fp2:
        return Fp2(PyFelt(1, self.p), PyFelt(0, self.p))

    def random(self) -> Fp2:
        return Fp2.random(self.p)

    @property
    def type(self) -> type[Fp2]:
        return Fp2


@dataclass(slots=True)
class ModuloCircuitElement:
    """
    Represents an element within a modulo circuit with its associated offset.

    Attributes:
        emulated_felt (PyFelt): The emulated field element in the modulo circuit
        offset (int): The offset in the values segment within the modulo circuit where the element is stored.
    """

    emulated_felt: PyFelt
    offset: int

    __repr__ = lambda self: f"ModuloCircuitElement({hex(self.value)}, {self.offset})"

    @property
    def value(self) -> int:
        return self.emulated_felt.value

    @property
    def p(self) -> int:
        return self.emulated_felt.p

    @property
    def felt(self) -> PyFelt:
        return self.emulated_felt


class Polynomial(Generic[T]):
    """
    Represents a polynomial with coefficients in a finite field.

    Parameters :
    coefficients (list[PyFelt | ModuloCircuitElement]): A list of coefficients for the polynomial.

    Magic Methods Summary:
    - __init__: Initializes a polynomial with a list of coefficients.
    - __add__: (Polynomial + Polynomial) Adds two polynomials.
    - __neg__: (-Polynomial) Negates the polynomial.
    - __sub__: (Polynomial - Polynomial) Subtracts one polynomial from another.
    - __mul__: (Polynomial * PyFelt|Polynomial) Multiplies the polynomial by either a field element or another polynomial.
    - __truediv__: (Polynomial / Polynomial) Divides one polynomial by another, asserting no remainder.
    - __floordiv__: (Polynomial // Polynomial) Performs floor division on two polynomials.
    - __mod__: (Polynomial % Polynomial) Finds the remainder of polynomial division.
    - __divmod__: Divides one polynomial by another, returning both quotient and remainder.
    - __eq__: (Polynomial == Polynomial) Checks if two polynomials are equal.
    - __neq__: (Polynomial != Polynomial) Checks if two polynomials are not equal.
    - __str__: Returns a string representation of the polynomial.
    - __pow__: (Polynomial ** int) Raises the polynomial to a power.
    """

    __slots__ = [
        "coefficients",
        "coeff_type",
        "p",
        "field",
        "zero_field",
        "zero_field_value",
    ]

    def __init__(self, coefficients: list[T], raw_init: tuple = None):
        if raw_init is not None:
            self._raw_init(coefficients, *raw_init)
        else:
            coeffs_types = {type(c) for c in coefficients}
            if coeffs_types == {PyFelt}:
                self._initialize(coefficients, PyFelt, BaseField)
            elif coeffs_types == {ModuloCircuitElement}:
                self._initialize([c.felt for c in coefficients], PyFelt, BaseField)
            elif coeffs_types == {Fp2}:
                self._initialize(coefficients, Fp2, BaseFp2Field)
            else:
                raise TypeError(
                    f"All elements in the list must be of the same type, either ModuloCircuitElement, PyFelt or Fp2., got {coeffs_types}"
                )

    def _raw_init(
        self,
        coefficients: list[T],
        coeff_type: type[T],
        p: int,
        field,
        zero_field,
        zero_field_value,
    ) -> None:
        self.coefficients = coefficients
        self.coeff_type = coeff_type
        self.p = p
        self.field = field
        self.zero_field = zero_field
        self.zero_field_value = zero_field_value

    def _initialize(
        self, coefficients: list[T], coeff_type: type[T], field_class: type
    ):
        self.coefficients: list[T] = coefficients
        self.coeff_type = coeff_type
        self.p = coefficients[0].p
        self.field = field_class(self.p)
        self.zero_field = self.field.zero()
        self.zero_field_value = self.zero_field.value

    def __repr__(self) -> str:
        if self.coeff_type == PyFelt:
            return f"Polynomial({[x.value for x in self.get_coeffs()]})"
        elif self.coeff_type == Fp2:
            return f"Polynomial({[x for x in self.get_coeffs()]})"

    def print_as_sage_poly(self, var_name: str = "z", as_hex: bool = False) -> str:
        """
        Prints the polynomial ready to be used in SageMath.
        """
        if self.is_zero():
            return ""
        coeffs = self.get_coeffs()
        string = ""
        zero = self.zero_field
        for i, coeff in enumerate(coeffs[::-1]):
            if coeff == zero:
                continue
            else:
                if self.coeff_type == PyFelt:
                    coeff_str = hex(coeff.value) if as_hex else str(coeff.value)
                elif self.coeff_type == Fp2:
                    coeff_str = f"({coeff.a1.value} * i + {coeff.a0.value})"

                if i == len(coeffs) - 1:

                    string += f"{coeff_str}"
                elif i == len(coeffs) - 2:
                    string += f"{coeff_str}*{var_name} + "
                else:
                    string += f"{coeff_str}*{var_name}^{len(coeffs) - 1 - i} + "
        return string

    def __getitem__(self, i: int) -> PyFelt:
        try:
            return self.coefficients[i]
        except IndexError:
            return self.zero_field

    def __len__(self) -> int:
        return len(self.coefficients)

    def degree(self) -> int:
        for i, coeff in enumerate(self.coefficients[::-1]):
            if coeff.value != self.zero_field_value:
                return len(self.coefficients) - 1 - i
        return -1

    def get_coeffs(self) -> list[T]:
        coeffs = self.coefficients[:]
        while len(coeffs) > 0 and coeffs[-1].value == self.zero_field_value:
            coeffs.pop()
        if coeffs == []:
            return [self.zero_field]
        return coeffs

    def get_value_coeffs(self) -> list[int]:
        if self.coeff_type == PyFelt:
            return [c.value for c in self.get_coeffs()]
        elif self.coeff_type == Fp2:
            raise NotImplementedError("Fp2 not implemented")

    def differentiate(self) -> "Polynomial":
        """
        Differentiates the polynomial with respect to x.

        Returns:
        Polynomial: The derivative of the polynomial.
        """
        if len(self.coefficients) <= 1:
            return Polynomial([self.zero_field])

        derivative_coeffs = [
            self.coefficients[i] * i for i in range(1, len(self.coefficients))
        ]
        return Polynomial(
            derivative_coeffs,
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )

    def __add__(self, other: Polynomial) -> Polynomial:
        if self.coeff_type != other.coeff_type:
            raise TypeError(
                f"Cannot add Polynomial of type {self.coeff_type} and {other.coeff_type} \n self: {self} \n other: {other}"
            )

        ns, no = len(self.coefficients), len(other.coefficients)
        if ns >= no:
            coeffs = self.coefficients[:]
            for i in range(no):
                coeffs[i] += other.coefficients[i]
        else:
            coeffs = other.coefficients[:]
            for i in range(ns):
                coeffs[i] += self.coefficients[i]

        return Polynomial(
            coeffs,
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )

    def __neg__(self) -> "Polynomial":
        return Polynomial(
            [-c for c in self.coefficients],
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )

    def __sub__(self, other: Polynomial) -> Polynomial:
        if self.coeff_type != other.coeff_type:
            raise TypeError(
                f"Cannot add Polynomial of type {self.coeff_type} and {other.coeff_type} \n self: {self} \n other: {other}"
            )

        ns, no = len(self.coefficients), len(other.coefficients)
        coeffs = self.coefficients[:]
        if ns >= no:
            for i in range(no):
                coeffs[i] -= other.coefficients[i]
        else:
            for i in range(ns):
                coeffs[i] -= other.coefficients[i]
            for i in range(ns, no):
                coeffs.append(-other.coefficients[i])

        return Polynomial(
            coeffs,
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )

    def __mul__(
        self, other: "Polynomial" | PyFelt | ModuloCircuitElement
    ) -> "Polynomial":
        if isinstance(other, (PyFelt, ModuloCircuitElement)):
            return Polynomial(
                [c * other.felt for c in self.coefficients],
                raw_init=(
                    self.coeff_type,
                    self.p,
                    self.field,
                    self.zero_field,
                    self.zero_field_value,
                ),
            )
        elif isinstance(other, Fp2):
            return Polynomial(
                [c * other for c in self.coefficients],
                raw_init=(
                    self.coeff_type,
                    self.p,
                    self.field,
                    self.zero_field,
                    self.zero_field_value,
                ),
            )
        if self.coeff_type != other.coeff_type:
            raise TypeError(
                f"Cannot multiply polynomial of type {self.coeff_type} by polynomial of type {other.type}"
            )

        len_self = len(self.coefficients)
        len_other = len(other.coefficients)
        buf = [self.zero_field] * (len_self + len_other - 1)
        for i in range(len_self):
            if self.coefficients[i].value == self.zero_field_value:
                continue  # optimization for sparse polynomials
            for j in range(len_other):
                buf[i + j] += self.coefficients[i] * other.coefficients[j]

        while len(buf) > 0 and buf[-1].value == self.zero_field_value:
            buf.pop()
        if buf == []:
            return Polynomial(
                [self.zero_field],
                raw_init=(
                    self.coeff_type,
                    self.p,
                    self.field,
                    self.zero_field,
                    self.zero_field_value,
                ),
            )

        return Polynomial(
            buf,
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )

    def __rmul__(
        self, other: "Polynomial" | PyFelt | ModuloCircuitElement
    ) -> "Polynomial":
        return self.__mul__(other)

    def __truediv__(self, other):
        quo, rem = Polynomial.__divmod__(self, other)
        assert (
            rem.is_zero()
        ), "cannot perform polynomial division because remainder is not zero"
        return quo

    def __floordiv__(self, other: "Polynomial") -> "Polynomial":
        quo, _ = Polynomial.__divmod__(self, other)
        return quo

    def __mod__(self, other: "Polynomial") -> "Polynomial":
        _, rem = Polynomial.__divmod__(self, other)
        return rem

    def __divmod__(self, denominator: "Polynomial") -> tuple[Polynomial, Polynomial]:
        den_deg = denominator.degree()
        if den_deg == -1:
            raise ValueError("Cannot divide by zero polynomial")
        num_deg = self.degree()
        if num_deg < den_deg:
            return (Polynomial.zero(self.p, self.coeff_type), self)

        remainder = Polynomial(
            self.coefficients[:],
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )
        quotient_coefficients = [self.zero_field] * (num_deg - den_deg + 1)

        denom_lead_inv = denominator[den_deg].__inv__()

        rem_deg = num_deg
        while rem_deg >= den_deg:
            shift = rem_deg - den_deg
            coefficient = remainder[rem_deg] * denom_lead_inv
            quotient_coefficients[shift] = coefficient

            subtractee = (
                Polynomial(
                    [self.zero_field] * shift + [coefficient],
                    raw_init=(
                        self.coeff_type,
                        self.p,
                        self.field,
                        self.zero_field,
                        self.zero_field_value,
                    ),
                )
                * denominator
            )
            remainder = remainder - subtractee
            rem_deg = remainder.degree()

        quotient = Polynomial(
            quotient_coefficients,
            raw_init=(
                self.coeff_type,
                self.p,
                self.field,
                self.zero_field,
                self.zero_field_value,
            ),
        )
        return quotient, remainder

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Polynomial):
            raise TypeError(f"Cannot compare Polynomial with {type(other)}")

        if self.degree() != other.degree():
            return False

        if self.degree() == -1:
            return True

        return self.get_coeffs() == other.get_coeffs()

    def __neq__(self, other: object) -> bool:
        return not self.__eq__(other)

    def leading_coefficient(self) -> PyFelt:
        return self.coefficients[self.degree()]

    def is_zero(self) -> bool:
        if not self.coefficients:
            return True
        for c in self.coefficients:
            if c != self.zero_field:
                return False
        return True

    @staticmethod
    def zero(p: int, type: type[T] = PyFelt) -> "Polynomial[T]":
        if type == PyFelt:
            return Polynomial([PyFelt(0, p)])
        elif type == Fp2:
            return Polynomial([Fp2.zero(p)])
        else:
            raise ValueError(f"Unknown type {type}")

    @staticmethod
    def one(p: int, type: type[T] = PyFelt) -> "Polynomial[T]":
        if type == PyFelt:
            return Polynomial([PyFelt(1, p)])
        elif type == Fp2:
            return Polynomial([Fp2.one(p)])
        else:
            raise ValueError(f"Unknown type {type}")

    def evaluate(self, point: PyFelt | Fp2) -> PyFelt | Fp2:
        assert isinstance(
            point, self.coeff_type
        ), f"point type must match polynomial type {self.coeff_type}"
        xi = self.field.one()
        value = self.zero_field
        for c in self.coefficients:
            value = value + c * xi
            xi = xi * point
        return value

    def __pow__(self, exponent: int) -> "Polynomial":
        if exponent == 0:
            return Polynomial([self.field.one()])
        acc = Polynomial([self.field.one()])
        for i in reversed(range(len(bin(exponent)[2:]))):
            acc = acc * acc
            if (1 << i) & exponent != 0:
                acc = acc * self
        return acc

    def pow(self, exponent: int, modulo_poly: "Polynomial") -> "Polynomial":
        if self.coeff_type != modulo_poly.coeff_type:
            raise TypeError(
                f"Cannot pow polynomial of type {self.coeff_type} modulo a polynomial of type {modulo_poly.coeff_type}"
            )
        one = Polynomial.one(self.p, self.coeff_type)
        if exponent == 0:
            return one
        acc = one
        for i in reversed(range(len(bin(exponent)[2:]))):
            acc = acc * acc % modulo_poly
            if (1 << i) & exponent != 0:
                acc = (acc * self) % modulo_poly
        return acc % modulo_poly

    def inv(self, modulo_poly: "Polynomial") -> "Polynomial":
        """
        Inverts a polynomial modulo another polynomial over a finite field.

        Parameters:
        modulo_poly Polynomial: The polynomial to invert modulo.

        Returns:
        Polynomial: The inverted polynomial.
        """
        if self.is_zero():
            raise ValueError("Cannot invert zero polynomial")
        inv, _, gcd = Polynomial.xgcd(self, modulo_poly)
        assert (
            gcd.degree() == 0 and gcd.coefficients[0] == 1
        ), f"Polynomial {self} is not invertible modulo {modulo_poly}"
        return inv

    @staticmethod
    def xgcd(x: Polynomial, y: Polynomial) -> tuple[Polynomial, Polynomial, Polynomial]:
        """
        Extended Euclidean Algorithm for polynomials.

        This method computes the extended greatest common divisor (GCD) of two polynomials x and y.
        It returns a tuple of three elements: (a, b, g) such that a * x + b * y = g, where g is the
        greatest common divisor of x and y. This is particularly useful in contexts like
        computational algebra or number theory where the coefficients of the polynomials are in a field.

        Parameters:
        x (Polynomial): The first polynomial.
        y (Polynomial): The second polynomial.

        Returns:
        tuple: A tuple (a, b, g) where:
            a (Polynomial): A polynomial such that a * x + b * y = g.
            b (Polynomial): A polynomial such that a * x + b * y = g.
            g (Polynomial): The greatest common divisor of x and y.
        """
        one = Polynomial.one(x.p, x.coeff_type)
        zero = Polynomial.zero(x.p, x.coeff_type)
        old_r, r = (x, y)
        old_s, s = (one, zero)
        old_t, t = (zero, one)

        while not r.is_zero():
            quotient = old_r // r
            old_r, r = (r, old_r - quotient * r)
            old_s, s = (s, old_s - quotient * s)
            old_t, t = (t, old_t - quotient * t)

        lcinv = old_r.leading_coefficient().__inv__()

        # a, b, g
        raw_init = (x.coeff_type, x.p, x.field, x.zero_field, x.zero_field_value)
        return (
            Polynomial([c * lcinv for c in old_s.coefficients], raw_init=raw_init),
            Polynomial([c * lcinv for c in old_t.coefficients], raw_init=raw_init),
            Polynomial([c * lcinv for c in old_r.coefficients], raw_init=raw_init),
        )

    @staticmethod
    def lagrange_interpolation(
        p: int, domain: list[PyFelt], values: list[PyFelt]
    ) -> Polynomial:
        """
        Performs Lagrange interpolation on a set of points.

        Parameters:
        p (int): The prime modulus for the field.
        domain (list[PyFelt]): The domain of the interpolation.
        values (list[PyFelt]): The values corresponding to the domain.

        Returns:
        Polynomial: The interpolated polynomial.
        """
        assert len(domain) == len(
            values
        ), "number of elements in domain does not match number of values -- cannot interpolate"
        assert len(domain) > 0, "cannot interpolate between zero points"
        field = BaseField(p)
        X = Polynomial([field.zero(), field.one()])
        acc = Polynomial([field.zero()])
        for i in range(len(domain)):
            prod = Polynomial([values[i]])
            for j in range(len(domain)):
                if j == i:
                    continue
                prod = (
                    prod
                    * (X - Polynomial([domain[j]]))
                    * Polynomial([(domain[i] - domain[j]).__inv__()])
                )
            acc = acc + prod
        return acc


@dataclass(slots=True)
class RationalFunction(Generic[T]):
    numerator: Polynomial[T]
    denominator: Polynomial[T]

    @property
    def field(self) -> BaseField | BaseFp2Field:
        return self.numerator.field

    @classmethod
    def zero(cls, p: int, type: type[T] = PyFelt) -> "RationalFunction[T]":
        return cls(Polynomial.zero(p, type), Polynomial.one(p, type))

    @classmethod
    def one(cls, p: int, type: type[T] = PyFelt) -> "RationalFunction[T]":
        return cls(Polynomial.one(p, type), Polynomial.one(p, type))

    def simplify(self) -> "RationalFunction":
        _, _, gcd = Polynomial.xgcd(self.numerator, self.denominator)
        num_simplified = self.numerator // gcd
        den_simplified = self.denominator // gcd
        return RationalFunction(
            num_simplified * self.denominator.leading_coefficient().__inv__(),
            den_simplified * den_simplified.leading_coefficient().__inv__(),
        )

    def __add__(self, other: "RationalFunction") -> "RationalFunction":
        return RationalFunction(
            self.numerator * other.denominator + other.numerator * self.denominator,
            self.denominator * other.denominator,
        ).simplify()

    def __mul__(self, other: int | PyFelt) -> "RationalFunction":
        if isinstance(other, int):
            other = self.field(other)
        elif isinstance(other, PyFelt):
            other = self.field(other.value)
        else:
            raise TypeError(f"Cannot multiply RationalFunction with {type(other)}")
        return RationalFunction(self.numerator * other, self.denominator)

    def evaluate(self, x: PyFelt | Fp2) -> PyFelt | Fp2:
        return self.numerator.evaluate(x) / self.denominator.evaluate(x)

    def degrees_infos(self) -> dict[str, int]:
        return {
            "numerator": self.numerator.degree(),
            "denominator": self.denominator.degree(),
        }


@dataclass(slots=True)
class FunctionFelt(Generic[T]):
    # f = a(x) + yb(x)
    a: RationalFunction[T]
    b: RationalFunction[T]

    @property
    def field(self) -> BaseField | BaseFp2Field:
        return self.a.numerator.field

    @classmethod
    def zero(cls, p: int, type: type[T] = PyFelt) -> "FunctionFelt[T]":
        return cls(RationalFunction.zero(p, type), RationalFunction.zero(p, type))

    @classmethod
    def one(cls, p: int, type: type[T] = PyFelt) -> "FunctionFelt[T]":
        return cls(RationalFunction.one(p, type), RationalFunction.zero(p, type))

    def simplify(self) -> "FunctionFelt":
        return FunctionFelt(self.a.simplify(), self.b.simplify())

    def __add__(self, other: "FunctionFelt") -> "FunctionFelt":
        return FunctionFelt(self.a + other.a, self.b + other.b)

    def __mul__(self, other: PyFelt | int) -> "FunctionFelt":
        return FunctionFelt(self.a * other, self.b * other)

    def __rmul__(self, other: PyFelt | int) -> "FunctionFelt":
        return self.__mul__(other)

    def evaluate(self, x: PyFelt | Fp2, y: PyFelt | Fp2) -> PyFelt | Fp2:
        assert (
            isinstance(x, self.field.type) and x.p == self.field.p
        ), f"x type must match field {self.field.type}, got {type(x)} over {hex(x.p)}"
        assert (
            isinstance(y, self.field.type) and y.p == self.field.p
        ), f"y type must match field {self.field.type}, got {type(y)} over {hex(y.p)}"

        return self.a.evaluate(x) + y * self.b.evaluate(x)

    def degrees_infos(self) -> dict[str, dict[str, int]]:
        return {
            "a": self.a.degrees_infos(),
            "b": self.b.degrees_infos(),
        }

    def validate_degrees(self, msm_size: int, batched: bool = True) -> bool:
        degrees = self.degrees_infos()
        if batched:
            extra = 2
        else:
            extra = 0
        assert degrees["a"]["numerator"] <= msm_size + 1 + extra
        assert degrees["a"]["denominator"] <= msm_size + 2 + extra
        assert degrees["b"]["numerator"] <= msm_size + 2 + extra
        assert degrees["b"]["denominator"] <= msm_size + 5 + extra
        return True

    def print_as_sage_poly(self, var: str = "x", as_hex: bool = False) -> str:
        return f"(({self.b.numerator.print_as_sage_poly(var, as_hex)}) / ({self.b.denominator.print_as_sage_poly(var, as_hex)}) * y + ({self.a.numerator.print_as_sage_poly(var, as_hex)} / ({self.a.denominator.print_as_sage_poly(var, as_hex)})"



import garaga.modulo_circuit_structs as structs
from garaga.definitions import CurveID
from garaga.precompiled_circuits.compilable_circuits.base import (
    BaseModuloCircuit,
    ModuloCircuit,
    PyFelt,
)
from garaga.precompiled_circuits.isogeny import IsogenyG1


class ApplyIsogenyCircuit(BaseModuloCircuit):
    def __init__(
        self, curve_id: int, auto_run: bool = True, compilation_mode: int = 1
    ) -> None:
        super().__init__(
            name=f"apply_isogeny_{CurveID(curve_id).name.lower()}",
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        return [self.field(44), self.field(4)]

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:

        circuit = IsogenyG1(self.name, self.curve_id, self.compilation_mode)

        px, py = circuit.write_struct(structs.G1PointCircuit(name="pt", elmts=input))

        x_affine, y_affine = circuit.run_isogeny(px, py)

        circuit.extend_struct_output(
            structs.G1PointCircuit(name="res", elmts=[x_affine, y_affine])
        )

        return circuit
